import jwt

from pydantic import BaseModel, ValidationError
from fastapi import Depends, status, HTTPException
from fastapi.security import HTTPBearer
from fastapi.security.http import HTTPAuthorizationCredentials

from .settings import settings

user_bearer = HTTPBearer(scheme_name="Ampel API token")

class User(BaseModel):

    name: str
    orgs: list[str]
    teams: list[str]

    @property
    def identities(self) -> list[str]:
        return [self.name] + self.orgs + self.teams


async def get_user(auth: HTTPAuthorizationCredentials = Depends(user_bearer)) -> User:
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(
            auth.credentials,
            settings.jwt_secret_key,
            algorithms=[settings.jwt_algorithm],
        )
        try:
            token_data = User(**payload)
            if not settings.allowed_identities.intersection(token_data.identities):
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)
            return token_data
        except ValidationError:
            raise credentials_exception
    except jwt.PyJWTError:
        raise credentials_exception



import json

from ampel.core.AmpelContext import AmpelContext
from fastapi import Depends, FastAPI, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from functools import cache
from pydantic import ValidationError

from .settings import settings
from .job import render_job, compact_json
from .auth import get_user, User
from .models import ArgoJobModel
from . import api

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import httpx

app = FastAPI(
    title="Argo job service",
    version="0.1.0",
    root_path=settings.root_path,
)


@cache
def get_context():
    return AmpelContext.load(
        config=settings.ampel_config,
        freeze_config=True,
    )


@app.post("/jobs/lint")
def lint_job(job: ArgoJobModel) -> dict:
    """
    Check a job template for errors
    """
    try:
        return render_job(get_context(), job)
    except ValidationError as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(exc.errors()),
        )


def render_template(
    job: ArgoJobModel,
    template: dict = Depends(lint_job),
    user: User = Depends(get_user),
) -> dict:
    return {
        "template": {
            "metadata": {
                "name": job.name,
                "labels": {
                    "ampelproject.github.io/creator": user.name,
                },
                "annotations": {
                    "ampelproject.github.io/job": job.json(),
                },
            },
            **template,
        }
    }


@app.get("/jobs")
async def list_jobs(user: User = Depends(get_user)):
    """
    Get all job templates
    """
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        (
            response := await client.get(
                f"api/v1/workflow-templates/{settings.namespace}",
                params={"listOptions.labelSelector": "ampelproject.github.io/creator"},
            )
        ).raise_for_status()
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()["items"]


@app.post("/jobs")
async def submit_job(template: dict = Depends(render_template)):
    """
    Submit a job template
    """
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        response = await client.post(
            f"api/v1/workflow-templates/{settings.namespace}",
            json=template,
        )
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()


@app.delete("/jobs/{name}")
async def delete_job(name: str, user: User = Depends(get_user)):
    """
    Delete an existing job template
    """
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        response = await client.delete(
            f"api/v1/workflow-templates/{settings.namespace}/{name}",
        )
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()


@app.put("/jobs/{name}")
async def update_job(name: str, template: dict = Depends(render_template)):
    """
    Update an existing job template
    """
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        (
            response := await client.get(
                f"api/v1/workflow-templates/{settings.namespace}/{name}"
            )
        ).raise_for_status()
        template["metadata"] = response.json()["metadata"]
        response = await client.put(
            f"api/v1/workflow-templates/{settings.namespace}/{name}",
            json=template,
        )
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()


@app.post("/jobs/{name}/rerender")
async def rerender_job(name: str, user: User = Depends(get_user)):
    """
    Re-render an existing job template
    """
    settings = api.KubernetesSettings.get()
    async with api.api_client() as client:
        (
            response := await client.get(
                f"api/v1/workflow-templates/{settings.namespace}/{name}"
            )
        ).raise_for_status()
        job_def = json.loads(
            response.json()["metadata"]["annotations"]["ampelproject.github.io/job"]
        )
        # work around legacy jobs where annotation was wrapped in an extra layer
        # of json serialization
        job = ArgoJobModel.parse_obj(
            json.loads(job_def) if isinstance(job_def, str) else job_def
        )
        template = {
            "template": {
                "metadata": response.json()["metadata"],
                **render_job(get_context(), job)
            }
        }
        response = await client.put(
            f"api/v1/workflow-templates/{settings.namespace}/{name}",
            json=template,
        )
    if response.status_code >= status.HTTP_400_BAD_REQUEST:
        raise HTTPException(status_code=response.status_code, detail=response.json())
    return response.json()


# If we are mounted under a (non-stripped) prefix path, create a potemkin root
# router and mount the actual root as a sub-application. This has no effect
# other than to prefix the paths of all routes with the root path.
if settings.root_path:
    wrapper = FastAPI()
    wrapper.mount(settings.root_path, app)
    app = wrapper



import os
import secrets
from typing import Set, Optional

import yaml
from pydantic import (
    BaseSettings,
    Field,
)


class Settings(BaseSettings):
    root_path: str = Field("", env="ROOT_PATH")
    ampel_config: str = Field("", env="AMPEL_CONFIG")
    ampel_image: str = Field(
        "gitlab.desy.de:5555/jakob.van.santen/docker-ampel:v0.8", env="AMPEL_IMAGE"
    )
    ampel_secrets: str = Field("ampel-secrets", env="AMPEL_SECRETS")
    image_pull_secrets: list[str] = Field(
        ["desy-gitlab-registry"], env="IMAGE_PULL_SECRETS"
    )
    job_env: list[dict] = Field(
        [
            {
                "name": "AMPEL_CONFIG_resource.mongo",
                "valueFrom": {
                    "secretKeyRef": {
                        "name": "mongo-live-admin-superuser",
                        "key": "connectionString.standard",
                    }
                },
            }
        ],
        env="JOB_ENV",
    )
    service_account: str = Field("argo-workflow", env="SERVICE_ACCOUNT")
    pod_priority_class: Optional[str] = Field(None, env="POD_PRIORITY_CLASS")

    jwt_secret_key: str = Field(secrets.token_urlsafe(64), env="JWT_SECRET_KEY")
    jwt_algorithm: str = Field("HS256", env="JWT_ALGORITHM")
    allowed_identities: Set[str] = Field(
        {"AmpelProject", "ZwickyTransientFacility"},
        env="ALLOWED_IDENTITIES",
        description="Usernames, teams, and orgs allowed to create workflow templates",
    )

    class Config:
        env_file = ".env"


try:
    with open(os.environ.get("AMPEL_ARGO_CONFIG", "/var/run/ampel-argo/config.yaml")) as f:
        config = yaml.safe_load(f)
except FileNotFoundError:
    config = {}

settings = Settings(**config)



import ast
from typing import Any
from ampel.config.AmpelConfig import AmpelConfig
from ampel.model.UnitModel import UnitModel
from ampel.model.job.JobModel import (
    InputArtifact,
    InputParameter,
    JobModel,
    TaskUnitModel,
    TemplateUnitModel,
    OutputParameter,
    InputArtifactHttpSource,
    ExpandWithItems,
    ExpandWithSequence,
)
from ampel.core.AmpelContext import AmpelContext
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate


from importlib import import_module
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from contextlib import contextmanager
from functools import singledispatch
import json

from typing import Union

from .settings import settings

# things that might be settable
# job template: will use ampel-job
# - image
# - channel
# - alias
# - extra parameters (also appears in step definition)
# - extra artifacts
# - env vars
# - secret mounts
# spec:
# - extra parameters (that may also appear in job template)
# - volumes (for secrets)
# - image pull secrets


class ExpressionTransformer(ast.NodeVisitor):
    """Translate top-level names"""

    def __init__(self, name_mapping: dict[str, str]):
        self._name_mapping = name_mapping
        self._output = ""

    def visit_Name(self, node: ast.Name):
        super().generic_visit(node)
        self._output += self._name_mapping.get(node.id, node.id)

    def visit_Attribute(self, node: ast.Attribute):
        super().generic_visit(node)
        self._output += "." + node.attr

    @classmethod
    def transform(cls, expression: str, name_mapping: dict[str, str]):
        self = cls(name_mapping)
        self.visit(ast.parse(expression, mode="eval"))
        return self._output


def translate_expression(expression: str) -> str:
    return (
        "{{ "
        + ExpressionTransformer.transform(
            expression, name_mapping={"job": "workflow", "task": "steps"}
        )
        + " }}"
    )


@singledispatch
def to_argo(model) -> dict:
    return JobModel.transform_expressions(model.dict(), translate_expression)


@to_argo.register # type: ignore[arg-type]
def _(model: list) -> list[dict]:
    return [to_argo(element) for element in model]


@to_argo.register
def _(model: OutputParameter) -> dict:
    return JobModel.transform_expressions(
        {
            "name": model.name,
            "valueFrom": {
                "path": str(model.value_from.path),
                "default": model.value_from.default,
            },
        },
        translate_expression,
    )


@to_argo.register
def _(model: ExpandWithItems):
    return {"withItems": model.items}


@to_argo.register
def _(model: ExpandWithSequence):
    return {"withSequence": model.sequence.dict()}


@to_argo.register
def _(model: InputArtifact) -> dict:
    return JobModel.transform_expressions(
        (
            model.dict()
            | {
                "path": str(model.path),
            }
        ),
        translate_expression,
    )


def get_template_for_task(
    job: JobModel,
    task: Union[TaskUnitModel, TemplateUnitModel],
    image="gitlab.desy.de:5555/jakob.van.santen/docker-ampel:v0.8",
) -> dict[str, Any]:
    return {
        "name": task.title,
        "inputs": {
            "parameters": [
                {"name": "name"},
            ]
            + to_argo(task.inputs.parameters), # type: ignore[operator]
            "artifacts": [
                {
                    "name": "__task",
                    "path": "/config/task.yml",
                },
                {
                    "name": "__channel",
                    "path": "/config/channel.yml",
                    "raw": {"data": compact_json(job.channel)},
                },
                {
                    "name": "__alias",
                    "path": "/config/alias.yml",
                    "raw": {"data": compact_json(job.alias)},
                },
            ]
            # explicitly avoid setting source parameters for input artifacts
            # to allow expressions to be expanded at step level
            + [
                {"name": artifact.name, "path": str(artifact.path)}
                for artifact in task.inputs.artifacts
            ],
        },
        "outputs": {
            "parameters": to_argo(task.outputs.parameters),
            "artifacts": to_argo(task.outputs.artifacts),
        },
        "metadata": {},
        "container": {
            "name": "main",
            "image": image,
            "command": [
                "ampel",
                "process",
                "--config",
                "/opt/env/etc/ampel.yml",
                "--secrets",
                "/config/secrets/secrets.yaml",
                "--channel",
                "/config/channel.yml",
                "--alias",
                "/config/alias.yml",
                "--db",
                "{{workflow.parameters.db}}",
                "--schema",
                "/config/task.yml",
                "--name",
                "{{inputs.parameters.name}}",
            ],
            "env": settings.job_env,
            "resources": {},
            "volumeMounts": [
                {
                    "name": "secrets",
                    "readOnly": True,
                    "mountPath": "/config/secrets",
                }
            ],
        },
    }


def get_unit_model(task: TaskUnitModel) -> dict[str, Any]:
    """get dict representation of UnitModel from TaskUnitModel"""
    return task.dict(exclude={"title", "expand_with", "inputs", "outputs"})


def render_task_template(ctx: AmpelContext, model: TemplateUnitModel) -> TaskUnitModel:
    """
    Resolve and validate a full AbsEventUnit config from given template
    """
    if model.template not in ctx.config._config["template"]:
        raise ValueError(f"Unknown process template: {model.template}")

    fqn = ctx.config._config["template"][model.template]
    class_name = fqn.split(".")[-1]
    Tpl = getattr(import_module(fqn), class_name)
    if not issubclass(Tpl, AbsProcessorTemplate):
        raise ValueError(f"Unexpected template type: {Tpl}")

    tpl = Tpl(**model.config)

    return TaskUnitModel(
        **(
            tpl.get_model(ctx.config._config, model.dict()).dict()
            | {
                "title": model.title,
                "expand_with": model.expand_with,
            }
        )
    )


@contextmanager
def job_context(ctx: AmpelContext, job: JobModel):
    """Add custom channels and aliases defined in the job"""
    old_config = ctx.config
    try:
        config = AmpelConfig(old_config.get(), freeze=False)
        config_dict = config._config
        for c in job.channel:
            dict.__setitem__(config_dict["channel"], str(c["channel"]), c)

        for k, v in job.alias.items():
            if "alias" not in config_dict:
                dict.__setitem__(config_dict, "alias", {})
            for kk, vv in v.items():
                if k not in config_dict["alias"]:
                    dict.__setitem__(config_dict["alias"], k, {})
                dict.__setitem__(config_dict["alias"][k], kk, vv)
        config.freeze()
        ctx.config = config
        ctx.loader.config = config
        yield ctx
    finally:
        ctx.config = old_config
        ctx.loader.config = old_config


compact_json = json.JSONEncoder(separators=(",", ":")).encode


def render_job(context: AmpelContext, job: JobModel):
    """Render Ampel job into an Argo workflow template spec"""

    steps = []
    templates = []

    with job_context(context, job) as ctx:
        for num, task_def in enumerate(job.task):

            task = (
                render_task_template(ctx, task_def)
                if isinstance(task_def, TemplateUnitModel)
                else task_def
            )
            # always raise exceptions
            if task.override is None:
                task.override = {}
            task.override["raise_exc"] = True

            with ctx.loader.validate_unit_models():
                try:
                    unit: UnitModel = UnitModel(**get_unit_model(task))
                except TypeError as exc:
                    # extract and raise inner ValidationError
                    raise exc.args[0] from None
                except ValueError as exc:
                    # wrap ValueError from missing unit
                    raise ValidationError([ErrorWrapper(exc, "unit")], model=UnitModel) from None

            if not "AbsEventUnit" in ctx.config._config["unit"][task.unit]["base"]:
                raise ValidationError(
                    [[ValueError(f"{task.unit} is not a subclass of AbsEventUnit")]],
                    model=JobModel,
                )

            templates.append(
                get_template_for_task(
                    job,
                    task,
                    image=settings.ampel_image,
                )
            )

            sub_step = {
                "name": task.title,
                "template": task.title,
                "arguments": {
                    "parameters": [{"name": "name", "value": task.title}],
                    "artifacts": [
                        {
                            "name": "__task",
                            "raw": {
                                "data": compact_json(
                                    job.transform_expressions(
                                        unit.dict(),
                                        translate_expression,
                                    )
                                )
                            },
                        },
                    ]
                    # drop path, as argo will ignore it in a step spec
                    + [(d, d.pop("path"))[0] for d in to_argo(task.inputs.artifacts)],
                },
            }
            if task.expand_with:
                sub_step |= to_argo(task.expand_with)

            steps.append([sub_step])

    workflow_template = {
        "spec": {
            "templates": templates
            + [
                {
                    "name": "workflow",
                    "inputs": {},
                    "outputs": {},
                    "metadata": {},
                    "steps": steps,
                },
            ],
            "entrypoint": "workflow",
            "arguments": {
                "parameters": [
                    {"name": "name", "value": job.name},
                    {"name": "db", "value": job.name},
                ]
                + [p.dict() for p in job.parameters]
            },
            "serviceAccountName": settings.service_account,
            "volumes": [
                {"name": "secrets", "secret": {"secretName": settings.ampel_secrets}}
            ],
            "imagePullSecrets": [{"name": n} for n in settings.image_pull_secrets],
        },
    }

    if settings.pod_priority_class:
        workflow_template["spec"]["podPriorityClassName"] = settings.pod_priority_class

    return workflow_template


def entrypoint():
    from argparse import ArgumentParser, FileType
    from ampel.core.AmpelContext import AmpelContext
    from .models import ArgoJobModel
    import yaml

    parser = ArgumentParser()
    parser.add_argument("--config", default="ampel.yaml")
    parser.add_argument("job", type=FileType("r"))

    args = parser.parse_args()
    ctx = AmpelContext.load(args.config)

    model = ArgoJobModel(**yaml.safe_load(args.job))

    resource = {
        "apiVersion": "argoproj.io/v1alpha1",
        "kind": "WorkflowTemplate",
        "metadata": {"name": model.name},
    } | render_job(ctx, model)

    print(yaml.dump(resource, sort_keys=False))
